// //////////////////////////////////////////////////////////////////////////
//
//     prob_dist.h
//     (c) 2007 Frank DiMaio
//
// //////////////////////////////////////////////////////////////////////////
//
//     This file is part of ACMI.
//
//     ACMI is free software; you can redistribute it and/or modify
//     it under the terms of the GNU General Public License as published by
//     the Free Software Foundation; either version 3 of the License, or
//     (at your option) any later version.
//
//     ACMI is distributed in the hope that it will be useful,
//     but WITHOUT ANY WARRANTY; without even the implied warranty of
//     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//     GNU General Public License for more details.
//
//     You should have received a copy of the GNU General Public License
//     along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// ////////////////////////////////////////////////////////////////////////////

#ifndef _PROB_DIST_H_
#define _PROB_DIST_H_ 1

#include <math.h>
#include <climits>
#include <cstdlib>
#include <vector>
#include <sstream>

using namespace std;

inline double square(double x) { return (x*x); }


// variance of a vector of doubles
double variance(vector<double> neigh_i) {
	double sum=0.0, sum2=0.0;
	int N=neigh_i.size();
	for (int i = 0; i < N; i++) {
		sum += neigh_i[i]; sum2 += neigh_i[i]*neigh_i[i];
	}
	return ( (N*sum2 - sum*sum)/ (N*N) );
}

// mean of a vector of doubles
double mean(vector<double> neigh_i) {
	double sum=0.0,N=neigh_i.size();
	for (int i = 0; i < N; i++) {
		sum += neigh_i[i];
	}
	return ( sum/N );
}

// mean of a vector of vector doubles
double variance(vector< vector<double> > neigh_i) {
	double sum=0.0, sum2=0.0;
	int Ni=neigh_i.size(), Nj, N=0;
	for (int i = 0; i < Ni; i++) {
		Nj = neigh_i[i].size();
		for (int j = 0; j < Nj; j++) {
			sum += neigh_i[i][j];
			sum2 += neigh_i[i][j]*neigh_i[i][j];
		}
		N += Nj;
	}
	return ( (N*sum2 - sum*sum)/ (N*N) );
}

// mean of a vector of vector of doubles
double mean(vector< vector<double> > neigh_i) {
	double sum=0.0;
	int Ni=neigh_i.size(), Nj, N=0;
	for (int i = 0; i < Ni; i++) {
		Nj = neigh_i[i].size();
		for (int j = 0; j < Nj; j++) {
			sum += neigh_i[i][j];
		}
		N += Nj;
	}
	return ( sum / N );
}

// normalized Gaussian PDF
inline double normCDF(double x) {
	return (.5*erfc(-x/sqrt(2.0)));
}
inline double normPDF(double x, double mean, double stdev) {
	return (1/(sqrt(2.0*M_PI)*stdev))*exp(-pow((x-mean),2)/(2*stdev*stdev));
}
inline double normPDF(double x) {
	return normPDF(x,0.0,1.0);
}

inline double normCDF(double x, double mean, double stdev) {
	return normCDF( (x-mean)/stdev );
}

// returns a normally distributed deviate with zero mean and unit
// variance, using ran1(idum) as the source of uniform deviates.
//
// from numerical recipes in C
double randnorm() {
	double fac,rsq,v1,v2;
	static int iset = 0;
	static double gset = 0.0;

	if (iset == 0) {
		do {
			v1=2.0*((double)rand()/RAND_MAX)-1.0; // pick two uniform numbers in the square extending from -1 to +1 in each direction,
			v2=2.0*((double)rand()/RAND_MAX)-1.0;
			rsq=v1*v1+v2*v2; // see if they are in the unit circle,
		} while (rsq >= 1.0 || rsq == 0.0); // and if they are not, try again.

		fac=sqrt(-2.0*log(rsq)/rsq); // Now make the Box-Muller transformation to get two normal deviates
		                             // Return one, save the other
		gset=v1*fac;
		iset = 1;
		return v2*fac;
	} else {
		iset = 0;
		return gset;
	}
}




// base class
class ProbDist {
public:
	virtual double getPDF(double x)=0;
	virtual double distr_stdev() = 0;
	virtual double distr_mean() = 0;
};

// histogram-based distribution
class HistogramDist : public ProbDist {
public:
	double getPDF(double x) {
		int bucket   = (int)floor((x-min)/step) + 1;

		// check if out of range
		if (bucket < 1)      bucket = 0;
		if (bucket > nSteps) bucket = nSteps+1;
		double retval =  ((double)counts[bucket]/(nObs)) / step ;

		//cout << "HistogramDist::getPDF(" << x << ") = " << retval << "\n";

		return (retval);
	}

	void setRange(double min, double max, int nSteps, int M=1) {
		this->M = M;
		this->min = min;
		this->nSteps = nSteps;
		this->step = (max-min)/nSteps;

		counts.clear();
		for (int i=0; i<nSteps+2; i++) counts.push_back(M);
		this->nObs = M*(nSteps+2);
	}

	void clear() {
		counts.clear();
		for (int i=0; i<nSteps+2; i++) counts.push_back(M);
		this->nObs = M*(nSteps+2);
	}

	inline void addObservation(double X) {
		//cout << "HistogramDist::addObservation(" << X << ")\n";

		int bucket = (int)floor((X-min)/step) + 1;
		if (bucket < 1)      bucket = 0;
		if (bucket > nSteps) bucket = nSteps+1;
		counts[bucket]++;
		nObs++;
	}

	void addObservations(vector<double> Xs) {
		int n = Xs.size();
		for (int i=0; i<n; i++)
			addObservation(Xs[i]);
	}

	double distr_stdev() {
		double sum=0, sum2=0;
		for (int i=0; i<=nSteps; i++) {
			sum += ( min+(i-0.5)*step ) * counts[i];
			sum2 += square( ( min+(i-0.5)*step ) ) * counts[i];
		}
		sum /= nObs;
		sum2 = sqrt(sum2/nObs - sum*sum);
		return (sum2);
	}

	double distr_mean() {
		double sum=0, sum2=0;
		for (int i=0; i<=nSteps; i++) {
			sum  += ( min+(i-0.5)*step ) * counts[i];
		}
		sum /= nObs;
		return (sum);
	}

	string to_string() {
		std::ostringstream oss;

		///////////////
		// stats
		//
		oss << " " << distr_mean() << " , " << distr_stdev() ;

		return oss.str();
	}

private:
	int M;   // M estimate
	double min, step;
	vector<int> counts;
	int nObs, nSteps;
};


// sum-of-gaussians distribution
class SumOfGaussiansDist : public ProbDist {
public:
	double getPDF(double x) {
		int n = Xs.size();
		if (n != stdevs.size()) computeStdevs();
		double retval = 0;

		for (int i=0; i<n; i++)
			retval += normPDF(x, Xs[i], stdevs[i]);
		retval /= (double)n;
		//cout << "SumOfGaussiansDist::getPDF(" << x << ") = " << retval << "\n";
		return (retval);
	}

	inline void addObservation(double X) {
		//cout << "SumOfGaussiansDist::addObservation(" << X << ")\n";

		Xs.push_back(X);
	}

	void addObservations(vector<double> Xins) {
		int n = Xins.size();
		for (int i=0; i<n; i++)
			addObservation(Xins[i]);
	}

	string to_string() {
		int n = Xs.size();
		std::ostringstream oss;
		oss << " " << mean(Xs) << " , " << sqrt(variance(Xs)) ;
		oss << " , s=" << (1.06*sqrt(variance(Xs))*pow(n,0.2));
		oss << " , [" << floor(Xs[0]+0.5);
		for (int i=1; i<n; i++) oss << "," << floor(Xs[i]+0.5);
		oss << "]";
		return oss.str();
	}

	double distr_stdev() {
		return (sqrt(variance(Xs)));
	}

	double distr_mean() {
		return (mean(Xs));
	}

private:
	void computeStdevs() {
		int n = Xs.size();
		stdevs.clear();

		// use Silverman '86 "rule-of-thumb"
		if (Xs.size() >= 3) {
			double S = 1.06 * sqrt(variance(Xs)) * pow(n,0.2);
			stdevs.resize(n,S);
		} else {
			stdevs.resize(n,300.0);
		}
	}

	vector<double> Xs;
	vector<double> stdevs;
};

#endif
