/*
 *  **************************************************************************
 *
 *  rmsd.h
 *  (c) 2005 Bosco K Ho
 *
 *  Implementation of the Kabsch algorithm to find the least-squares
 *  rotation matrix for a superposition between two sets of vectors.
 *  The jacobi transform of the diagonalization of a symmetric matrix
 *  is taken from Numerical Recipes. This piece of code is
 *  self-contained and *does not require* another library.
 *
 *  **************************************************************************
 *
 *  Modifications (c) 2007 Frank DiMaio 
 *
 *  **************************************************************************
 *
 *     This file is part of ACMI.
 * 
 *     ACMI is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 3 of the License, or
 *     (at your option) any later version.
 * 
 *     ACMI is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  **************************************************************************
 *
 */

#ifndef _rmsd_h_
#define _rmsd_h_

/*
 * calculate_rotation_rmsd_notranslate()
 *
 *   given two XYZ lists, computes the rmsd and  rotation
 *      of the optimal superpostion.
 *   rotation is about the origin (0,0,0)
 *
 *   modified FPD, 12/05
 */
void calculate_rotation_rmsd_notranslate(
                             double ref_xlist[][3],
                             double mov_xlist[][3],
                             int n_list,
                             double U[3][3],
                             double* rmsd);

/*
 * calculate_rotation_rmsd()
 *
 *   given two lists of x,y,z coordinates, constructs
 *    - mov_com: the centre of mass of the mov list
 *    - mov_to_ref: vector between the com of mov and ref
 *    - U: the rotation matrix for least-squares, usage of
 *         of the matrix U[3][3] is
 *           for (i=0; i<3; i++)
 *           {
 *             rotated_v[i] = 0.0;
 *             for (j=0; j<3; j++)
 *               rotated_v[i] += U[i][j] * v[j];
 *           }
 *    - rmsd: measures similarity between the vectors
 */
void calculate_rotation_rmsd(double ref_xlist[][3],
                             double mov_xlist[][3],
                             int n_list,
                             double mov_com[3],
                             double mov_to_ref[3],
                             double U[3][3],
                             double* rmsd);

/*
 * fast_rmsd()
 *
 * Fast calculation of rmsd w/o calculating a rotation matrix,
 * adapted from the BTK by Chris Saunders 11/2002.
 */
void fast_rmsd(double ref_xlist[][3],
               double mov_xlist[][3],
               int n_list,
               double* rmsd);


#endif
