// //////////////////////////////////////////////////////////////////////////
//
//     shfear.cpp
//     (c) 2007 Frank DiMaio
//
// //////////////////////////////////////////////////////////////////////////
//
//     This file is part of ACMI.
//
//     ACMI is free software; you can redistribute it and/or modify
//     it under the terms of the GNU General Public License as published by
//     the Free Software Foundation; either version 3 of the License, or
//     (at your option) any later version.
//
//     ACMI is distributed in the hope that it will be useful,
//     but WITHOUT ANY WARRANTY; without even the implied warranty of
//     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//     GNU General Public License for more details.
//
//     You should have received a copy of the GNU General Public License
//     along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// //////////////////////////////////////////////////////////////////////////

#include <time.h>    // benchmarking

#include <sys/stat.h>  // mkdir
#include <sys/types.h>

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

extern "C" {
#include "fftw3.h"
#include "complex.h"
#include "csecond.h"
#include "so3_correlate_fftw.h"
#include "soft_fftw_pc.h"
#include "soft_fftw.h"
#include "makeWigner.h"
#include "makeweights.h"
#include "FST_semi_memo.h"
#include "cospmls.h"
#include "primitive_FST.h"
#include "seminaive.h"
}

#include "spharm.h"
#include "getopts.h"   // parse command line args

int POSITIONAL_COST[] = {1,1,1,1,1};  // allow ANY position to change
int MAX_PAMDIST = 20;
const int WIDTH = 2; //number of grid points, in each direction, to allocate probability for a partial solution.

//
// main()
//

int main(int argc, char** argv) {
	srand(1);
	/////////////////////
	/////////////////////
	///
	/// read input files
	Options shfearOpts;
  	int switchInt;

	//                  Short   Long        Description           Take args?
	shfearOpts.addOption("w",    "working",      "[REQUIRED] Working Directory",      true);
	shfearOpts.addOption("m",    "map",          "[REQUIRED] Input fs4 map file",      true);
	shfearOpts.addOption("c",    "cryst",        "[REQUIRED] Input crystal file",      true);
	shfearOpts.addOption("s",    "seq",          "[REQUIRED] Input sequence file",      true);
	shfearOpts.addOption("r",    "reso",          "Density map resolution (default 2.0)",      true);
	shfearOpts.addOption("B",    "bandwidth",     "Spherical harmonic bandwidth (default 12)",      true);
	shfearOpts.addOption("R",    "radius",        "Fragment radius (default 5)",      true);
	shfearOpts.addOption("N",    "nfrags",        "Number of fragments considered per amino acid (default 50)",      true);
	shfearOpts.addOption("x",    "grid",          "Approximate map grid spacing (default 1A)",      true);
	shfearOpts.addOption("o",    "omit",          "Using density values, omit fraction of map (default 0.8)",     true);
	shfearOpts.addOption("p",    "parallel",      "Parallelize <job#> of <Njobs>",      true);
	shfearOpts.addOption("a",    "analogues",     "List of structural analogues for fragment generation",      true);
	shfearOpts.addOption("f",    "force",         "Force priors to be overwritten",     false);
	shfearOpts.addOption("P",    "partial",       "Partial solution",     true);
	shfearOpts.addOption("W",    "width",         "Width of Gaussian kernel on partial solution, in grid units (default 1)", true);
	shfearOpts.addOption("v",    "verbose",       "Verbose output for debugging", false);
	shfearOpts.addOption("e",    "seleniums",     "Substitute sulfurs with seleniums in Methionine residues", false);
	shfearOpts.addOption("A",    "augment",       "Repeat shfear, but expand previous results based on given directory", true);

	bool verbose = false;
	string workingdir, mapfile, crystfile, seqfile, seleniumfile, augmentfile;
	vector< string >  similar_files;
	vector< CAtrace > similar_pdbs;
	bool forceOverwrite = false;
	CAtrace priorStruct;

	// defaults
	double grid=1.0, CC_CUTOFF=1.0, percentile=0.8, seleniumprior=0.9, sigma = 1.0;
	int beginAtRes=0, skipByRes=1, Rmax=5, nfrags=50, nSimilar=0;
	Bout = B = 12;
	reso = 2.0;

	shfearOpts.parse(argc, argv);
	while ((switchInt = shfearOpts.cycle()) >= 0) {
		vector<string> pargs;
		Crystal nullCryst;

		switch(switchInt) {
		case 0:   // Option 'working-dir'
			workingdir =  shfearOpts.getArgs(switchInt); break;
		case 1:   // Option 'map'
			mapfile =  shfearOpts.getArgs(switchInt); break;
		case 2:   // Option 'cryst'
			crystfile =  shfearOpts.getArgs(switchInt); break;
		case 3:   // Option 'seq'
			seqfile =  shfearOpts.getArgs(switchInt); break;
		case 4:   // Option 'reso'
			reso = atof( shfearOpts.getArgs(switchInt) ); break;
		case 5:   // Option 'bandwidth'
			B = Bout = atoi( shfearOpts.getArgs(switchInt) ); break;
		case 6:   // Option 'radius'
			Rmax = atoi( shfearOpts.getArgs(switchInt) ); break;
		case 7:   // Option 'nfrags'
			nfrags = atoi( shfearOpts.getArgs(switchInt) ); break;
		case 8:   // Option 'grid'
			grid = atof( shfearOpts.getArgs(switchInt) ); break;
		case 9:   // Option 'omit'
			percentile = atof( shfearOpts.getArgs(switchInt) ); break;
		case 10:   // Option 'parallel'
			Tokenize( shfearOpts.getArgs(switchInt) , pargs );
			if (pargs.size() == 2) {
				skipByRes = atoi(pargs[1]);
				if(skipByRes < 1){
					cerr << "Error, improper use of -p parameter.  Second number must be greater than zero" << endl;
					exit(1);
				}
				beginAtRes = pos_mod( atoi(pargs[0])-1 , skipByRes );
			}
			break;
		case 11:  // Option 'analogues'
			Tokenize( shfearOpts.getArgs(switchInt) , similar_files );
			nSimilar = similar_files.size();
			for (int i=0; i<nSimilar; ++i)
				similar_pdbs.push_back( CAtrace( similar_files[i] , nullCryst ) );
			break;
		case 12:  // Option 'force'
			forceOverwrite = true;
			break;
		case 13:
			priorStruct =  CAtrace(shfearOpts.getArgs(switchInt),nullCryst,true); break;
		case 14:
			sigma = atof( shfearOpts.getArgs(switchInt) ); break;
			if(sigma <= 0) {
				cerr << "Warning: illegal Gaussian width "<< sigma << ", must be greater than 0.  Setting to 0.1" << endl;
				sigma = 0.1; //prevent Nan values
			}
			if(sigma > 3){
				cerr << "Warning: Gaussian width "<< sigma << " does not provide enough signal for partial solution. Setting to 3" << endl;
				sigma =3;
			}
		case 15:
			verbose = true;
		case 16: //Option 'seleniums'
			SUB_SULFURS = true; //Global variable in protein.h
		case 17: // Option 'augment'
					augmentfile =  shfearOpts.getArgs(switchInt); break;
		}
	}

	// at a minimum, need to specify workingdir, map, cryst, seq
	if (workingdir.length() == 0 || mapfile.length() == 0 || crystfile.length() == 0 || seqfile.length() == 0) {
		cerr << "Error: Must specify a working directory, fs4 map file, crystal file, and sequence file.\n";
		cerr << "       See \'" << argv[0] << " -h\' for more details\n";
		exit(1);
	}


	map_3d<double> origMap;
	Crystal crystData;
	readDensityMap(origMap,mapfile,crystData,crystfile);
	Sequence seq(seqfile);

	// density map grid
	Rstep = 1.0;
	vec3<int> dims = crystData.getDims(grid);
	int dimsProduct = dims[0]*dims[1]*dims[2];
	crystData.buildMaps(dims);

	// resample map
	map_3d<double> map(dims);
	resize_fft(origMap, map);

	// output directories
	string outdir = workingdir+"/priors/";
	//string outdir = workingdir+"/priors.shfear"+itoc(B)+"/";
	mkdir(workingdir.c_str(), 0755);   // ensure dirs exists
	mkdir(outdir.c_str(),     0755);

	//////////////////////
	///
	///  pretty-print
	if (skipByRes > 1) {
		cerr << "*** PARALLEL RUN #" << beginAtRes+1 << " of " << skipByRes << " *** " << endl;
		cerr << endl;
	}
	cerr << "Running with inputs: " << endl;
	cerr << "    working directory = \'" << workingdir << "\'" << endl;
	cerr << "    fs4-format map    = <" << mapfile << ">" << endl;
	cerr << "    crystal dims      = " << crystData.to_string() << endl;
	cerr << "            spacegp   = " << crystData.symmGpId << endl;
	cerr << "    protein sequence  = " << seq.to_string() << endl;
	cerr << "    resolution        = " << reso << endl;

	cerr << endl;
	cerr << "Search parameters " << endl;
	cerr << "    B                 = " << B << endl;
	cerr << "    radius            = " << Rmax << endl;
	cerr << "    nfrags            = " << nfrags << endl;

	cerr << endl;
	if (nSimilar > 0) {
		cerr << "Using similar maps: " << endl;
		for (int i=0; i<nSimilar; ++i)
			cerr << "    " << similar_files[i] << endl;
	}

	// only consider the top n%  locations
	percentile = min(1.0,percentile);
	percentile = max(percentile,0.0);

	vector<double> projection = map.project(crystData.getAsymmUnit());
	sort(projection.begin(), projection.end());
	double thresh = projection[(int)floor(percentile*projection.size()) - 1];
	map_3d<double> filter(map);
	cerr << "Unit cell grid <" << dims.to_string() << ">" << endl;
	cerr << "    In asymm unit : " << projection.size() << endl;
	cerr << "    With density > " << thresh << "(=" << floor(100*percentile+0.5) << " %ile): "
	     << ceil((1-percentile)*projection.size()) << endl;
	cerr << endl;

	// map spherical harmonics
	map_3d<double> sigR(Rmax,2*B,2*B);
	map_3d<double> sigCoefR(Rmax,B,B);
	map_3d<double> sigCoefI(Rmax,B,B);

	map_3d< cubicSplineCoeff3d > sphPtrsMap(Rmax,2*B,2*B);
	cache_spherical_resampling_cubic_spline(sphPtrsMap, dims, crystData);

	// build tiny unit cells
	//    for constructing maps from fragments
	//    (we don't want to build over the entire crystal)
	double radiusSmall = Rmax * Rstep + 1.0;
	double cellEdgeSmall = 2.0 * radiusSmall + 12.0;
	int nRadsteps;
	double deltaX = min(1.0, reso/3.0);

	// small -- for DB lookup
	nRadsteps = (int)ceil( (cellEdgeSmall) / deltaX );
	Crystal crystSmall(cellEdgeSmall,cellEdgeSmall,cellEdgeSmall, 90.00,90.00,90.00);
	vec3<int> dimsSmall(nRadsteps, nRadsteps, nRadsteps);
	map_3d< cubicSplineCoeff3d > sphPtrsSmall(Rmax,2*B,2*B);
	cache_spherical_resampling_cubic_spline(sphPtrsSmall, dimsSmall, crystSmall);

	SOFT_initialize();
	//
	///////////////////

	///////////////////
	//
	//  for each residue, consider region centered on that residue
	//      cluster all instances of that 5mer, saving centroids
	//      extract spherical harmonics
	if(augmentfile.length()>0){
				cerr << "Augmenting previous run" << endl;
	}
	for (int i_res = beginAtRes; i_res < seq.nres(); i_res+=skipByRes) {
		vector< Nmer >               centroids;
		vector< map_3d<double> >     coeffR_centroids;
		vector< map_3d<double> >     coeffI_centroids;
		vector< double >             vars;

		bool prior = false;
		if (seq.serialGet(i_res) == '_') continue;  // gap


		//Is this defined in prior structure

		string mer = seq.serialGet5mer(i_res);

		string outfile = outdir + "match_" + mer;
		if(priorStruct.nres() > i_res && priorStruct.serialGet(i_res).typeShort != '_'){
			prior = true;
			outfile = outfile + itoc(i_res);
		}
		string infile = augmentfile+"match_" + mer;
		if(priorStruct.nres() > i_res && priorStruct.serialGet(i_res).typeShort != '_'){
			infile = infile + itoc(i_res);
		}


		if(augmentfile.length()==0){
			if (exists( string(outfile+".cc.mat")) && exists( string(outfile+".betaF.mat")) &&
					exists( string(outfile+".betaB.mat")) && exists(string(outfile+".gammaB.mat")) &&
					exists( string(outfile+".gammaF.mat"))) {
				if (!forceOverwrite) {
					cerr << outfile << " exists, continuing (use '-f' to force overwrite)\n";
					continue;
				} else {
					cerr << outfile << " exists, will be overwritten\n";
				}
			}
		} else {
			if (!(exists( string(infile+".cc.mat")) && exists( string(infile+".betaF.mat")) &&
								exists( string(infile+".betaB.mat")) && exists(string(infile+".gammaB.mat")) &&
								exists( string(infile+".gammaF.mat")))){
				cerr << "Error: match file does not exist for augmenting fragment " << mer << endl;
				exit(1);
			}
		}

		//
		// do DB lookup and clustering to find nmer
		//
		// compute SphHarmonics for each centroid fragment
		map_3d< orientation > matchOrient( dims , orientation(-10,-10,-10,-10) );  // outside normal range of beta, gamma
		map_3d< double >      matchMax( dims, 0 );

		vec3<int> X;
		int sumAsymm = (int)crystData.getAsymmUnit().sum(), counter = 0;

		double CCsum=0.0, CCsum2=0.0;
		int nCC=0;
		if(!prior){ //do not run if residue is partial solution and no mixture is required.
		if(verbose) cerr << "Searching for templates" << endl;
		// first, lookup in file
		if (nSimilar > 0) {
			spharm_lookupFromFilelist(mer, similar_files, similar_pdbs, radiusSmall, dimsSmall, crystSmall, sphPtrsSmall,
			                          centroids, coeffR_centroids, coeffI_centroids);
		}
		// then, ONLY IF no nmers are found in the PROVIDED pdb file
		//    should we go to the database
		if (centroids.size() == 0) {
			spharm_lookup(mer, nfrags, radiusSmall, dimsSmall, crystSmall, sphPtrsSmall,
			              centroids, coeffR_centroids, coeffI_centroids);
		}

		int nfragments = centroids.size();
		if (nfragments == 0) { cerr << "WARNING! No clusters found for \'" << mer << "\'.\n"; continue; }
		cerr << "Searching density map for " << nfragments
		     << " templates corresponding to sequence \'" << mer << "\'" << endl;

		if(augmentfile.length()>0){
			matchMax.readFromFile(infile+".cc.mat");
			readMapsFromDisk(matchOrient,infile);
		}
		for (X[0]=0; X[0]<dims[0]; ++X[0]) {
			for (X[1]=0; X[1]<dims[1]; ++X[1]) {
				for (X[2]=0; X[2]<dims[2]; ++X[2]) {
					if (crystData.getAsymmUnit().get(X[0],X[1],X[2]) == 0) continue;
					if (filter.get(X[0],X[1],X[2]) < thresh) continue;
					if(matchMax.get(X[0],X[1],X[2]) > 0) continue; //already searched (augment mode)
					// 1) extract
					spherical_resample_cubic_spline(sigR, map, sphPtrsMap, X);

					// 2) standardize
					sph_standardize(sigR);

					// 3) spharm
					SOFT_sharm_transform(sigR, sigCoefR, sigCoefI);

					// 4) compare to each template fragment
					for (int i_frag=0; i_frag < nfragments; i_frag++) {
						Nmer &frag = centroids[i_frag];
						vec3<double> CAminus1 = (mer[1] == '_' ? vec3<double>( 3.8,0.0,0.0) : frag.getCA(1));
						vec3<double> CAplus1  = (mer[3] == '_' ? vec3<double>(-3.8,0.0,0.0) : frag.getCA(3));

						SOFT_so3_correlate(sigCoefR,sigCoefI, coeffR_centroids[i_frag],coeffI_centroids[i_frag]);

						// find best rotation, (correlation) for THIS cluster
						double maxval = -INFINITY, thisVal;
						int maxloc = -1;
						for (int i=0 ; i<8*Bout*Bout*Bout ; ++i) {
							thisVal = so3SigSum.get(i).real();
							CCsum += thisVal;
							CCsum2 += thisVal*thisVal;
							nCC++;
							if (thisVal >= maxval) {
								maxval = thisVal; maxloc = i ;
							}
						}
						// now compare to other clusters
						if ( maxval > matchMax.get(X[0],X[1],X[2]) ) {
							matchMax.set(X[0],X[1],X[2],  maxval);

							mat33<double> R_abg;
							SOFT_idx_to_rot(maxloc,R_abg);

							vec3<double> R_CAm1 = CAminus1.apply(R_abg);
							vec3<double> R_CAp1 = CAplus1.apply(R_abg);

							if ( isnan(R_CAm1.beta()) || isnan(R_CAm1.gamma())
							|| isnan(R_CAp1.beta()) || isnan(R_CAp1.gamma()) ) {
								cerr << "ERROR! NaN in orientation\n";
								cerr << "     " << CAminus1.to_string() << " --> " << R_CAm1.to_string()
								       << " --> (" << R_CAm1.beta() << "," << R_CAm1.gamma() << ")" << endl;
								cerr << "     " << CAplus1.to_string() << " --> " << R_CAp1.to_string()
								       << " --> (" << R_CAp1.beta() << "," << R_CAp1.gamma() << ")" << endl;
								exit(1);
							}

							matchOrient.set(X[0],X[1],X[2],
							    orientation(R_CAm1.beta(),R_CAm1.gamma(),R_CAp1.beta(),R_CAp1.gamma()));
						}
					} // end loop over clusters
				}
			}
		}// end loop over map
		} else {
			cerr << "Using prior structure as template corresponding to sequence \'" << mer << "\'" << endl;
		}
		//If a prior structure is provided, set the match probabilitiy accordingly.
		if(prior){
			PDBResidue pres = priorStruct.serialGet(i_res); //map coord
			vec3<double> coord (pres.cartX.apply(crystData.c2f)); //frac coords
			vec3<double> dimsD(dims[0],dims[1],dims[2]);
			coord = coord*dimsD;
			coord.mod(dimsD);

			float mB(-10),mG(-10),pB(-10),pG(-10);
			if(i_res > 0 && priorStruct.serialGet(i_res-1).typeShort != '_'){
				vec3<double> CAminus1 = priorStruct.serialGet(i_res-1).cartX-pres.cartX;
				mB = CAminus1.beta();
				mG = CAminus1.gamma();
			}

			if(i_res < priorStruct.nres()-1 && priorStruct.serialGet(i_res+1).typeShort != '_'){
				vec3<double> CAplus1 = priorStruct.serialGet(i_res+1).cartX-pres.cartX;
				pB = CAplus1.beta();
				pG = CAplus1.gamma();
			}
			for(int i = (int)max(coord[0]-WIDTH,0.0); i <= min(coord[0]+WIDTH,dimsD[0]-1); i++){
				 for(int j = (int)max(coord[1]-WIDTH,0.0); j <= min(coord[1]+WIDTH,dimsD[1]-1); j++){
				 	for(int k = (int)max(coord[2]-WIDTH,0.0); k <= min(coord[2]+WIDTH,dimsD[2]-1); k++){
				 		matchMax.set(i,j,k,normPDF(i,coord[0],sigma)*normPDF(j,coord[1],sigma)*normPDF(k,coord[2],sigma)); //place Gaussian kernel on true solution
				 		matchOrient.set(i,j,k, orientation(mB,mG,pB,pG));
				 	}
				 }
			}
		}
		if(verbose) cerr << "Applying to symmetric copies..." ;
		//
		// symmetric transfer
		map_3d<int> ptrMap(dims), symMap(dims);
		crystData.maxSymmetric(matchMax, ptrMap, symMap);
		for (int j=0; j<dimsProduct; j++) {
			int jptr = ptrMap.get(j);
			if (jptr == j) continue;
			orientation &j_orient = matchOrient.get(j);
			orientation &jptr_orient = matchOrient.get(jptr);

			// forward
			vec3<double> x_F ( cos(jptr_orient.gammaF)*sin(jptr_orient.betaF),
			                   sin(jptr_orient.gammaF)*sin(jptr_orient.betaF),
			                   cos(jptr_orient.betaF) );
			x_F = x_F.apply(crystData.c2f);     // convert to fractional
			x_F = x_F.apply(crystData.getSymmInvRot( symMap.get(j)));  // rotate
			x_F = x_F.apply(crystData.f2c);     // convert back to cartesian

			// backward
			vec3<double> x_B ( cos(jptr_orient.gammaB)*sin(jptr_orient.betaB),
			                   sin(jptr_orient.gammaB)*sin(jptr_orient.betaB),
			                   cos(jptr_orient.betaB) );
			x_B = x_B.apply(crystData.c2f);     // convert to fractional
			x_B = x_B.apply(crystData.getSymmInvRot( symMap.get(j)));  // rotate
			x_B = x_B.apply(crystData.f2c);     // convert back to cartesian

			j_orient.betaF = (jptr_orient.betaF == -10) ? -10 : x_F.beta();
			j_orient.gammaF =  (jptr_orient.gammaF == -10) ? -10 : x_F.gamma();
			j_orient.betaB =  (jptr_orient.betaB == -10) ? -10 : x_B.beta();
			j_orient.gammaB =  (jptr_orient.gammaB == -10) ? -10 : x_B.gamma();
		}
		if(verbose) cerr << "done\n" ;
		//
		// write to disk
		string outfilestem = outdir + "match_" + mer;
		if(priorStruct.nres() > i_res && priorStruct.serialGet(i_res).typeShort != '_'){
			outfilestem += itoc(i_res);
		}
		writeMapToDisk (matchMax   , string(outfilestem + ".cc"));
		writeMapsToDisk(matchOrient, outfilestem);
		cerr << "Search over density map for \'" << mer << "\' ("<< i_res+1<<") complete\n" << endl;
	}  // end loop over residues
	cerr << "Run complete" << endl;
}
