// //////////////////////////////////////////////////////////////////////////
//
//     protein.h
//     (c) 2007 Frank DiMaio
//
// //////////////////////////////////////////////////////////////////////////
//
//     This file is part of ACMI.
//
//     ACMI is free software; you can redistribute it and/or modify
//     it under the terms of the GNU General Public License as published by
//     the Free Software Foundation; either version 3 of the License, or
//     (at your option) any later version.
//
//     ACMI is distributed in the hope that it will be useful,
//     but WITHOUT ANY WARRANTY; without even the implied warranty of
//     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//     GNU General Public License for more details.
//
//     You should have received a copy of the GNU General Public License
//     along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// //////////////////////////////////////////////////////////////////////////


#ifndef _PROTEIN_H
#define _PROTEIN_H 1

#include <map>
#include <algorithm>

#include "crystal.h"
#include "chem.h"
#include "protein_fk.h"

extern "C" {
	#include "rmsd.h"
}

bool SUB_SULFURS = false;  //Substitute sulfurs with seleniums in methionines
/////////////////////////
/////////////////////////
///
///  PDBResidue
///
class PDBResidue {
public:
	PDBResidue() {
		fracX = vec3<double>(0.0,0.0,0.0);
		cartX = vec3<double>(0.0,0.0,0.0);
		typeLong = "___";
		typeShort = '_';
	}

	vec3<double> fracX, cartX;
	int id;
	string typeLong;
	char typeShort;
	double B;
};

// no crystal symmetry
double getSquaredError(vec3<double> &A, vec3<double> &B, mat33<double> f2c) {
	vec3<double> delf = (A - B); delf.minmod1();
	return delf.apply(f2c).norm2();
}

// look at all symmetric copies
double getSquaredError(vec3<double> &A, vec3<double> &B, Crystal &crystData) {
	double thisdist, mindist;
	vec3<double> thisdelf, mindelf, symmFracX;
	int nsymm = crystData.nSymm();

	mindelf = (A - B); mindelf.minmod1();
	mindist = mindelf.apply(crystData.f2c).norm2();

	for (int j=1; j<nsymm; ++j) {
		crystData.getSymmetric(B, symmFracX, j);
		thisdelf = (A - symmFracX); thisdelf.minmod1();
		thisdist = thisdelf.apply(crystData.f2c).norm2();

		if (thisdist < mindist) {
			mindist = thisdist;
			mindelf = thisdelf;
		}
	}

	return mindist;
}

double getSquaredError(PDBResidue &A, PDBResidue &B, mat33<double> f2c) {
	return getSquaredError(A.fracX, B.fracX, f2c);
}

double getSquaredError(PDBResidue &A, PDBResidue &B, Crystal &crystData) {
	return getSquaredError(A.fracX, B.fracX, crystData);
}


/////////////////////////
/////////////////////////
///
///  PDBChain
///
class PDBChain{
public:
	PDBChain() {
		idChar = ' ';
	}

	PDBChain(char chnId) {
		idChar = chnId;
	}

	void clear() {
		reses.clear();
	}

	void setId(char id) {
		idChar = id;
	}

	void addResidue(PDBResidue R) {
		reses.push_back(R);
	}

	// add a residue at the specified index
	void addResidue(int idx, PDBResidue R) {
		if (idx >= reses.size())
			reses.resize(idx+1);

		reses[idx] = R;
	}
	void removeResidue() {
		reses.pop_back();
	}


	PDBResidue back() {
		return reses.back();
	}

	int nres() { return (reses.size()); }
	char getId()  { return idChar; }

	// return the likelihood gain of this chain
	// assumes likelihood of an individual residue is stored in the B factor
	// likelihood may be either be log-likelihood or real likelihood
	// prior is based on uniform distribution over unit cell and is usu. log(1/prod(dims))
	double likelihoodGain(double prior) {
		double gain = 0.0;
		for (int i=0; i<reses.size(); ++i) {
			gain += reses[i].B<0 ? reses[i].B : log(reses[i].B) - prior;
		}
		return gain;
	}

	string get5mer(int i) {
		std::ostringstream oss;
		oss << (char) (i-2<0 || i-2>=reses.size() ? '_' : reses[i-2].typeShort )
		    << (char) (i-1<0 || i-1>=reses.size() ? '_' : reses[i-1].typeShort )
		    << (char) (i  <0 || i  >=reses.size() ? '_' : reses[i].typeShort)
		    << (char) (i+1<0 || i+1>=reses.size() ? '_' : reses[i+1].typeShort)
		    << (char) (i+2<0 || i+2>=reses.size() ? '_' : reses[i+2].typeShort);
		return oss.str();
	}

	vector< vec3<double> > getFracCoords() {
		vector< vec3<double> > retval(reses.size());

		for (int i=0; i<reses.size(); i++)
			retval[i] = reses[i].fracX;

		return retval;
	}

	inline PDBResidue& operator[](int i) { return reses[i]; }

private:
	vector<PDBResidue> reses;
	char idChar;
};

/////////////////////////
/////////////////////////
///
///  CAtrace
///
class CAtrace {
public:
	CAtrace(){};

	CAtrace(string pdbfile, Crystal &crystData, bool USE_PDB_RESID = false, double minDist = INFINITY) {
		ifstream pdbin(pdbfile.c_str());
		if (!pdbin.is_open()) {
			cerr << "ERROR: Unable to open PDB file '" << pdbfile << "'\n";
			throw 0;
		}

		string line;
		bool inserted;
		PDBChain thisChain;
		PDBResidue thisRes;
		double partial  = 0, sumPartial = 0;

		while(getline(pdbin,line)) {
			if ( (line.substr(0,3) == "TER") ) {
				// new chain
				chains.push_back( thisChain );
				thisChain.clear();
			}

			if ( (line.substr(0,6) == "CRYST1" ) ) {
				crystData.init(line);
			}

			if (    (line.substr(0,6) == "ATOM  " || line.substr(0,6) == "HETATM")
				 && (line.substr(13,2) == "CA") ) {

				//Handle partial atom locations
				if(sumPartial >= 0.99){
					sumPartial = 0;
					partial = 0;
				}
				if(partial > 0.01){
					double thisPartial = atof(line.substr(56,4));
					sumPartial += thisPartial;
					if( thisPartial > partial){
						partial = thisPartial;
						thisChain.removeResidue();
					} else {
						continue;
					}
				} else {
					partial = 1.00 - atof(line.substr(56,4));
					sumPartial += partial;
				}


 				if (thisChain.nres() == 0) {
					thisChain.setId(line[21]);
				}


				// add a new residue to the current chain
				thisRes.cartX = vec3<double>(atof(line.substr(30,8)),
				                             atof(line.substr(38,8)),
				                             atof(line.substr(46,8)));
				thisRes.fracX = thisRes.cartX.apply(crystData.c2f); thisRes.fracX.mod(1);
				thisRes.id = atoi(line.substr(22,4).c_str());
				thisRes.typeLong = line.substr(17,3);
				thisRes.typeShort = oneLetterCode(thisRes.typeLong);
				thisRes.B = atof(line.substr(61,6));

				// is this a new chain?
				if (thisChain.nres() > 0) {
					if ( (thisRes.cartX - thisChain.back().cartX).norm2() > minDist*minDist
					     || thisChain.getId() != line[21] ) {
						chains.push_back(thisChain);
						thisChain.clear();
						thisChain.setId(line[21]);
					}
				}

				if (USE_PDB_RESID)
					thisChain.addResidue(thisRes.id-1, thisRes);  // -1 since PDB typ. starts at '1'
				else
					thisChain.addResidue(thisRes);

			}
		}

		// done reading the file, insert final chain in correct location
		if (thisChain.nres() > 0) {
			chains.push_back( thisChain );
		}
	}

	CAtrace(PDBChain& chain) {
		chains.push_back(chain);
	}

	void generateSymmetric(Crystal &crystData) {
		//
		// make symm copies of chain
		PDBResidue thisRes, symmRes, prevSymmRes;
		PDBChain thisChain;
		int nsymm = crystData.nSymm();
		int nchains = chains.size();
		for (int i=1; i<nsymm; ++i) {   // symmgp1 == identity
			for (int j=0; j<nchains; ++j) {
				thisChain.clear();
				thisChain.setId(chains[j].getId());
				for (int k=0; k<chains[j].nres(); ++k) {
					thisRes = chains[j][k];
					symmRes = thisRes;
					crystData.getSymmetric(thisRes.fracX, symmRes.cartX, i);

					if (k == 0) {
						symmRes.cartX = symmRes.fracX.apply(crystData.f2c);
					} else {
						vec3<double> delta = symmRes.fracX - prevSymmRes.fracX; delta.minmod1();
						symmRes.cartX = prevSymmRes.cartX + delta.apply(crystData.f2c);

					}
					prevSymmRes = symmRes;
					thisChain.addResidue(symmRes);
				}
				chains.push_back( thisChain );
			}
		}
	}

	int nChains() {
		return chains.size();
	}

	CAtrace getSubsequence(char chnID, int startAt, int nres) {
		PDBChain thischain(chnID);

		for (int i=0; i<chains.size(); i++) {
			if (chnID != chains[i].getId()) continue;
			for (int j=0; j<chains[i].nres(); ++j) {
				if (chains[i][j].id >= startAt && chains[i][j].id < startAt+nres) {
					thischain.addResidue(chains[i][j]);
				}
			}
		}

		cout << endl;

		return CAtrace(thischain);
	}

	vector< vec3<double> > getCoords() {
		vector< vec3<double> > coords;
		for (int i=0; i<chains.size(); i++)
			for (int j=0; j<chains[i].nres(); ++j)
				coords.push_back( chains[i][j].cartX );

		return coords;
	}

	PDBResidue getByResID(char chnID, int resID) {
		for (int i=0; i<chains.size(); i++) {
			if (chnID != chains[i].getId()) continue;
			for (int j=0; j<chains[i].nres(); ++j) {
				if (resID == chains[i][j].id)
					return chains[i][j];
			}
		}
		return PDBResidue();

		/*cerr << "CATrace::getByResID(" << chnID << "," << resID << ") error: res not found!\n";
		cerr << "    chain[0][0].resid = " << chains[0].getId() << "," << chains[0][0].id << endl;
		exit(1);*/
	}

	/////////////////////
	//
	// serialize all copies of all UNIQUE chains
	int nres() {
		int reses = 0;
		for (int i=0; i<chains.size(); i++)
			reses += chains[i].nres();
		return reses;
	}

	PDBResidue serialGet(int toGet) {
		for (int i=0; i<chains.size(); i++)
			if (toGet < chains[i].nres())
				return chains[i][toGet];
			else
				toGet -= chains[i].nres();
		cerr << "CATrace::serialGet(int) error!\n";  // out of range
		exit(1);
	}

	void serialGet(int toGet, int &resId, int &chnId) {
		for (int i=0; i<chains.size(); i++) {
			if (toGet < chains[i].nres()) {
				resId = toGet;
				chnId = i;
				return;
			} else {
				toGet -= chains[i].nres();
			}
		}
		resId = -1; chnId = -1;  // out of range
	}

	void serialGet(int toGet, int &resId, char &resType, char &chnId) {
		for (int i=0; i<chains.size(); i++) {
			if (toGet < chains[i].nres()) {
				//resId = toGet;
				resId   = chains[i][toGet].id;
				resType = chains[i][toGet].typeShort;
				chnId = chains[i].getId();
				return;
			} else {
				toGet -= chains[i].nres();
			}
		}
		resId = -1; resType = 'X'; chnId = ' ';  // out of range
	}

	vector<int> serialGetCopies(int toGet) {
		int resId = -1, toGetCounter=toGet;
		vector<int> sims;

		for (int i=0; i<chains.size(); i++) {
			if (toGetCounter < chains[i].nres()) {
				resId = toGetCounter;
				break;
			} else {
				toGetCounter -= chains[i].nres();
			}
		}
		if (resId < 0) return sims;

		int chainOffsetId = 0;
		string mer = serialGet5mer(toGet);
		for (int i=0; i<chains.size(); i++) {

			string mer2 = serialGet5mer(chainOffsetId+resId);

			if (    resId < chains[i].nres()
			     && toGet != chainOffsetId+resId
				 && mer2[2] != '_' && mer[2] != '_'
				 && pam_distance(mer, mer2) <= 1 ) {
				sims.push_back(chainOffsetId+resId);
				//cerr << "SIM " << toGet << " and " << chainOffsetId+resId << endl;
			}

			chainOffsetId += chains[i].nres();
		}

		return sims;
	}

	vector<int> getCopies(int chnId, int resId) {
		vector<int> sims;

		int chainOffsetId = 0;
		string mer = chains[chnId].get5mer(resId);
		for (int i=0; i<chains.size(); i++) {
			if (i == chnId) continue;
			if (resId >= chains[i].nres()) continue;

			string mer2 = chains[i].get5mer(resId);

			if (    mer2[2] != '_' && mer[2] != '_'
				 && pam_distance(mer, mer2) <= 1 ) {
				sims.push_back(i);
			}

			chainOffsetId += chains[i].nres();
		}

		return sims;
	}
	//

	// get subsequence; represent out-of-range residues as '_'
	string serialGet5mer(int toGet) {
		if (toGet < 0)
			return "_____";

		for (int i=0; i<chains.size(); i++)
			if (toGet < chains[i].nres())
				return chains[i].get5mer(toGet);
			else
				toGet -= chains[i].nres();
		// out of range ...
		return "_____";
	}

	string serialGetConcat5mer(int toGet) {
		string retval = "_____";

		for (int pos = 0; pos<5; pos++) {
			int thisToGet = toGet + pos - 2;
			for (int i=0; i<chains.size(); i++)
				if (thisToGet >= 0 && thisToGet < chains[i].nres()) {
					retval[pos] = chains[i][thisToGet].typeShort;
					break;
				} else
					thisToGet -= chains[i].nres();
		}
		// out of range ...
		return retval;
	}

	int getLongestChain(PDBChain& chain) {
		int longestChain = 0, bestChainIdx, bestResIdx;
		int currResIdx = 0;
		for (int i=0; i<chains.size(); ++i) {

			if (chains[i].nres() > longestChain) {
				longestChain = chains[i].nres();
				bestChainIdx = i;
				bestResIdx = currResIdx;
			}

			currResIdx += chains[i].nres();
		}
		chain = chains[bestChainIdx];
		return bestResIdx;
	}

	void getLongestChains(vector<PDBChain>& chains, vector<int>& startAts) {
		chains.clear(); startAts.clear();
		int currResIdx = 0;
		for (int i=0; i<chains.size(); ++i) {
			if (chains[i].nres() >= 4) {
				chains.push_back(chains[i]);
				startAts.push_back(currResIdx);
			}
			currResIdx += chains[i].nres();
		}
	}

	int getMLChain(PDBChain& chain, double logPrior) {
		int bestChainIdx, bestResIdx;
		int currResIdx = 0;
		double thisGain, bestGain=-INFINITY;
		for (int i=0; i<chains.size(); ++i) {
			thisGain = chains[i].likelihoodGain(logPrior);

			cout << "gain = " << thisGain << " at " << i << "," << currResIdx << endl;

			if (thisGain > bestGain) {
				bestGain = thisGain;
				bestChainIdx = i;
				bestResIdx = currResIdx;
			}

			currResIdx += chains[i].nres();
		}
		chain = chains[bestChainIdx];
		return bestResIdx;
	}
	PDBChain operator[](char chnID){
		if(chnID =='0') return chains[0];
		for (int i=0; i<chains.size(); i++) {
			if (chnID == chains[i].getId()) return chains[i];
		}
		cerr << "CATrace[](char) error!\n";  // out of range
		exit(1);
	}

	inline PDBChain& operator[](int i) { return chains[i]; }
	inline const PDBChain& operator[](int i) const { return chains[i]; }

private:
	vector< PDBChain >  chains;
};




/////////////////////////
/////////////////////////
///
///  Chain
///
class Chain{
public:
	Chain() { init(' '); }
	Chain(char chnId) { init(chnId); }
	Chain(char chnId, string chnSeq) { init(chnId, chnSeq); }
	Chain(char chnId, string chnSeq, string chnContext) { init(chnId, chnSeq, chnContext); }

	void init(char chnId) {
		idChar = chnId;
		seq = "";
		context = "____";
	}
	void init(char chnId, string chnSeq) {
		idChar = chnId;
		seq = chnSeq;
		context = "____";
	}
	void init(char chnId, string chnSeq, string chnContext) {
		idChar = chnId;
		seq = chnSeq;
		context = chnContext;
	}

	int    nres()   { return seq.size(); }
	char   getId()  { return idChar; }
	string getSeq() { return seq; }

	// get subsequence; represent out-of-range residues as '_'
	string get5mer(int i) {
		std::ostringstream oss;
		oss << (char) getAt(i-2) << (char) getAt(i-1)
		    << (char) getAt(i)
		    << (char) getAt(i+1) << (char) getAt(i+2);
		return oss.str();
	}
	string get5merContext(int i) {
		std::ostringstream oss;
		oss << (char) getAtContext(i-2) << (char) getAtContext(i-1)
		    << (char) getAtContext(i)
		    << (char) getAtContext(i+1) << (char) getAtContext(i+2);
		return oss.str();
	}
	Chain getSubsequence(int startAt, int length) {
		std::ostringstream seq_ss, ctxt_ss;
		ctxt_ss << getAt(startAt-2)      << getAt(startAt-1)
		        << getAt(startAt+length) << getAt(startAt+length+1);
		for (int i=startAt; i<startAt+length; ++i) {
			seq_ss << (char) getAt(i);
			//cout << "[" << (char) getAt(i) << "] ";
		}
		return Chain(idChar, seq_ss.str(), ctxt_ss.str());
	}

	inline char operator[](int i) { return seq[i]; }
	inline const char operator[] (int i) const { return seq[i]; }

	void set(int i, char restype) {
		if (seq.size() < i+1) {
			seq.resize(i+1, '_');
		}
		seq[i] = restype;
	}

	// access w/ bounds check
	char getAt(int i) {
		if (i >= 0 && i<seq.size())
			return seq[i];
		return '_';
	}
	char getAtContext(int i) {
		if (i >= 0 && i<seq.size())
			return seq[i];
		if (i == -1 || i == -2)
			return context[i+2];
		if (i==seq.size() || i == seq.size()+1)
			return context[i-seq.size()+2];
		return '_';
	}

private:
	string seq, context;
	char idChar;
};

/////////////////////////
/////////////////////////
///
///  Sequence
///
class Sequence {
public:
	Sequence() {}

	// construct a seq from a chain
	Sequence(Chain chn) {
		chains.push_back( vector<Chain>(1, chn) );
	}
	Sequence(string seqfile){
		load(seqfile);
	}
	// read a set of CA chains from a PDB file
	void load(string seqfile) {
		ifstream seqin(seqfile.c_str());
		if (!seqin.is_open()) {
			cerr << "ERROR: Unable to open sequence file '" << seqfile << "'\n";
			throw 0;
		}

		string line, seq;
		char chainId=('A'-1);
		bool inserted;

		while(getline(seqin,line)) {
			if(line[0] == '>') { //FASTA Format
				seqin.close();
				loadFASTA(seqfile);
				break;
			}
			if (line[1] == ':') {
				chainId = line[0];
				seq = line.substr(2);
			} else {
				chainId++;  // autogen
				seq = line;
			}

			chains.push_back( vector<Chain>(1, Chain(chainId,seq)) );
		}
	}
	void loadFASTA(string seqfile){
		ifstream seqin(seqfile.c_str());

		string line, seq;
		char chainId=('A'-1);

		while(getline(seqin,line)) {
			if (line[0] == '>') {
				chainId++;
				if(seq.length() != 0){
					chains.push_back( vector<Chain>(1, Chain(chainId,seq)) );
				}
			}
			seq += line;
		}
	}

	int nChains() { return chains.size(); };
	int nChainCopies() {
		int retval = 0;
		for (int i=0; i<chains.size(); i++) {
			retval += chains[i].size();
		}
		return retval;
	};

	inline vector<Chain>& operator[](int i) { return chains[i]; }
	inline const vector<Chain>& operator[](int i) const { return chains[i]; }

	/////////////////////
	//
	// serialize all copies of all UNIQUE chains
	int nres() {
		int reses = 0;
		for (int i=0; i<chains.size(); i++)
			reses += chains[i][0].nres();
		return reses;
	}

	void set(int chnid, int resid, char restype) {
		if (chains.size() <= chnid) {
			chains.resize(chnid+1);
			chains[chnid].push_back( Chain() );  // add an empty chain
		}
		chains[chnid][0].set(resid , restype);
	}

	void serialGet(int toGet, int &resId, int &chnId) {
		for (int i=0; i<chains.size(); i++) {
			if (toGet < chains[i][0].nres()) {
				resId = toGet;
				chnId = i;
				return;
			} else {
				toGet -= chains[i][0].nres();
			}
		}
		resId = -1; chnId = -1;  // out of range
	}

	int getSerialId(int resId, int chnId) {
		int idx=resId;
		for (int i=0; i<chnId; i++) {
			idx += chains[i][0].nres();
		}
		return idx;
	}
	void serialGet(int toGet, int &resId, char &resType, char &chnId) {
		for (int i=0; i<chains.size(); i++) {
			if (toGet < chains[i][0].nres()) {
				resId = toGet;
				resType = chains[i][0][toGet];
				chnId = chains[i][0].getId();
				return;
			} else {
				toGet -= chains[i][0].nres();
			}
		}
		resId = -1; resType = 'X'; chnId = ' ';  // out of range
	}

	char serialGet(int toGet) {
		for (int i=0; i<chains.size(); i++)
			if (toGet < chains[i][0].nres())
				return chains[i][0][toGet];
			else
				toGet -= chains[i][0].nres();
		return '_';  // out of range
	}

	int serialGetNCopies(int toGet) {
		for (int i=0; i<chains.size(); i++)
			if (toGet < chains[i][0].nres())
				return chains[i].size();
			else
				toGet -= chains[i][0].nres();
		return 0;  // out of range
	}

	// get subsequence; represent out-of-range residues as '_'
	string serialGet5mer(int toGet) {
		for (int i=0; i<chains.size(); i++)
			if (toGet < chains[i][0].nres())
				return chains[i][0].get5mer(toGet);
			else
				toGet -= chains[i][0].nres();
		// out of range ...
		return "_____";
	}
	string serialGet5merContext(int toGet) {
		for (int i=0; i<chains.size(); i++)
			if (toGet < chains[i][0].nres())
				return chains[i][0].get5merContext(toGet);
			else
				toGet -= chains[i][0].nres();
		// out of range ...
		return "_____";
	}

	Sequence serialGetSubsequence(int startAt, int length) {
		for (int i=0; i<chains.size(); i++) {
			if (startAt < chains[i][0].nres())
				return Sequence( chains[i][0].getSubsequence(startAt, length) );
			else
				startAt -= chains[i][0].nres();
		}

		// out of range
		return Sequence();
	}

	vector<int> serialGetCopies(int toGet) {
		int resId = -1, toGetCounter=toGet;
		vector<int> sims;
		for (int i=0; i<chains.size(); i++) {
			if (toGetCounter < chains[i][0].nres()) {
				resId = toGetCounter;
				break;
			} else {
				toGetCounter -= chains[i][0].nres();
			}
		}
		if (resId < 0) return sims;

		int chainOffsetId = 0;
		string mer = serialGet5mer(toGet);
		for (int i=0; i<chains.size(); i++) {
			string mer2 = serialGet5mer(chainOffsetId+resId);

			if (    resId < chains[i][0].nres()
				 && mer[2] != '_' && mer2[2] != '_'
				 && pam_distance(mer, mer2) <= 1 ) {
				sims.push_back(chainOffsetId+resId);
			}

			chainOffsetId += chains[i][0].nres();
		}

		return sims;
	}
	//
	/////////////////////

	string to_string() {
		std::ostringstream oss;
		for (int i=0; i<20; ++i)
			oss << (char)serialGet(i);
		if (nres() > 20)
			oss << "...";
		return oss.str();
	}

private:
	vector< vector< Chain > >  chains;
};






/////////////////////////
/////////////////////////
///
///  Atom
///
class Atom {
public:
	int elt;            // element
	vec3<double> cartX; // position in cartesian coords
	double B;           // temp factor
	string atmid;

	void parseLine(string line) {
		// location, B factor
		cartX = vec3<double>(atof(line.substr(30,8)),
							 atof(line.substr(38,8)),
							 atof(line.substr(46,8)));

		//B = 0.0;  // ignore B
		B = atof(line.substr(61,6));

		// element type
		//      Kind of a mess, BUT ... element names may be one or two letters,
		//      may start in either column 12 or 13, and may be ambiguous
		//        (calcium and C-alpha are both CA and only distinguished by
		//         starting col# -- 12 for calcium, 13 for C-alpha),
		//      so checks are made in a certain order.
		elt = eltLookup( line.substr(12,2) );
		if (elt == -1) {    // unknown atom type
			if (line.substr(13,3) == "UNK") {
				elt = 6;    // model unknown as C
			} else {
				elt = eltLookup( line.substr(13,1) );

				if (elt == -1) {    // unknown atom type
					elt = eltLookup( line.substr(12,1) );

					if (elt == -1) {    // unknown atom type
						elt = eltLookup( line.substr(13,2) );

						if (elt == -1)     // unknown atom type
							cerr << "UNKNOWN FRAGMENT ATOM >" << line.substr(12,4) << "<\n";
					}
				}
			}
		}
		atmid = line.substr(12,4);
	}

	//
	// serialize/unserialize this object to an ostream/from an istream
	void serialize(ostream &out) {
		float X;
		out.write((char*)&elt, sizeof(int));
		X = (float)cartX[0]; out.write((char*)&X, sizeof(float));
		X = (float)cartX[1]; out.write((char*)&X, sizeof(float));
		X = (float)cartX[2]; out.write((char*)&X, sizeof(float));
		X = (float) B;       out.write((char*)&X, sizeof(float));
	}

	void unserialize(istream &in) {
		float X;
		in.read((char*)&elt, sizeof(int));
		in.read((char*)&X, sizeof(float)); cartX[0] = X;
		in.read((char*)&X, sizeof(float)); cartX[1] = X;
		in.read((char*)&X, sizeof(float)); cartX[2] = X;
		in.read((char*)&X, sizeof(float)); B = X;
	}

	string to_string() {
		ostringstream out;
 		out << "[" << atmid << "] " << cartX.to_string() ;
		return out.str();
	}
};


/////////////////////////
/////////////////////////
///
///  Residue
///
class Residue {
public:
	Residue() {
		restype = 23;
		b = 0.0;
	}

	void rotate(mat33<double> R) {
		int natoms = atoms.size();
		for (int i=0; i<natoms; ++i)
			atoms[i].cartX = atoms[i].cartX.apply(R);
	}

	void translate(vec3<double> T) {
		int natoms = atoms.size();
		for (int i=0; i<natoms; ++i)
			atoms[i].cartX = atoms[i].cartX + T;
	}

	int restype;
	vec3<double> CAm1,  C,CA,N,O,  CAp1;  // cache neighboring backbone atoms
	double b;
	vector< Atom > atoms;             // list of all atoms (sidechain+backbone)
};


/////////////////////////
/////////////////////////
///
///  Nmer
///
class Nmer {
	friend void align   (Nmer &, Nmer &, mat33<double> &, double &, bool, bool );
	friend void calcRMSd(Nmer &, Nmer &, double &);
	friend void align   (vector< vec3<double> >, Nmer &, mat33<double> &, double &);
public:
	Nmer(int N=5) {
		merLength = N;
		extent = 0.0;
	}

	// as a fragment of a PDB file
	Nmer(string pdbfile, char chain, int id, double radius = 0.0, int N=5) {
		merLength = N;
		readFileId(pdbfile,chain,id, radius);
	}

	//
	// read an nmer contained w/i a PDB file, using actual PDB IDs
	void readFileId(string pdbfile, char tgtChain, int tgtResid, double radius = 0.0) {
		this->pdbfile = pdbfile;
		this->tgtChain = tgtChain;
		this->tgtResid = tgtResid;

		atoms.clear(); resids.clear(); atoms_sphere.clear(); center_res.clear();
		mass = extent = 0.0;

		double radius2 = square(radius+1.0);

		atoms.resize(merLength);//store all atoms of mer
		resids.resize(merLength, RESID_NONE); //store ids
		resTypes.resize(merLength);	// store residue types

		ifstream pdbin(pdbfile.c_str());	//open pdf file
		if (!pdbin.is_open()) {
			cerr << "ERROR: Unable to open PDB file '" << pdbfile << "'\n";
			throw 0;
		}

		vec3<double> centerCA;
		streampos fragStart;
		streampos fileSpot,lastSpot;
		Atom thisAtm;
		int thisResId=-1, lastResId = -1, breakResId;
		int Noffset = (merLength-1)/2;
		string line;
		bool fragRead = false;
		char thisChain;

		//
		// PASS 1
		//     find resids of included residues
		//     find coords of center CA
		//
		while(getline(pdbin,line)) {
			if(line.length() > 25){
				if(atoi(line.substr(22,4)) != lastResId){
					fileSpot = lastSpot;
				}
				lastResId = atoi(line.substr(22,4));
			}
			if (    (line.substr(0,6) == "ATOM  " || line.substr(0,6) == "HETATM")     // atom entry
			     && (tgtChain == '0' || tgtChain == line[21])                          // correct chain
			     && (line[16]=='A' || line[16]==' ')                                   // correct conformation
			     && (line.substr(13,2) == "CA")                                        // alpha carbon
			   ) {
				thisChain = line[21];
				thisResId = atoi(line.substr(22,4));

				if (    thisResId >= tgtResid - Noffset
				            && thisResId <= tgtResid + Noffset ) { //in fragment
					if ( !fragRead) {
						fragStart = fileSpot;//pdbin.tellg();
						fragRead = true;
					}
					int pos = thisResId - tgtResid + Noffset;
					resids[pos] = thisResId ;
					resTypes[pos] = line.substr(17,3);

					if ( (thisResId == tgtResid)  ) {
						centerCA = vec3<double>(atof(line.substr(30,8)),
												atof(line.substr(38,8)),
												atof(line.substr(46,8)));
						tgtChain = thisChain;
					}
				} else if(thisResId > tgtResid + Noffset){ // past fragment
					break;
				}
			}
			lastSpot = pdbin.tellg();
		}
		if(!fragRead){
			cerr << "Warning, did not find Nmer in "<< pdbfile << ", Chain "
				<< tgtChain << ", ResID " << tgtResid << endl;
			return;
		}

		//breakResId=resids[0]; for (int i=1; i<merLength; i++) breakResId=max(breakResId, resids[i]);
		//
		// PASS 2
		//     read in included residues
		//     translate coordinates so center CA is at the origin
		//
		pdbin.clear();   // clear eof()
		pdbin.seekg(0, ios::beg);
		fragRead = false;

		while(getline(pdbin,line)) {
			if (     (line.substr(0,6) == "ATOM  " || line.substr(0,6) == "HETATM")
			     && (tgtChain == line[21])
			     && ( line[16]=='A' || line[16]==' ' )
				 &&   line[13] != 'H' ) {
				thisResId = atoi(line.substr(22,4));
				int pos = thisResId - tgtResid + Noffset;

				// 5-mer
				if (pos >= 0 && pos < merLength && line.substr(17,3) != "HOH") {
					fragRead = true;

					thisAtm.parseLine(line);
					//Repalce sulfurs with seleniums If the center atom is a methionine and the user
					//specified flag is set
					if (thisAtm.elt == 16 && pos == merLength/2 && resTypes[pos] == "MET" && SUB_SULFURS){
						thisAtm.elt = 34;
					}
					thisAtm.cartX = thisAtm.cartX - centerCA;  // center fragment
					if (radius != 0.0 && thisAtm.cartX.norm2() <= radius2) {
						atoms_sphere.push_back(thisAtm);
						//cout << thisAtm.to_string() << endl;
					}
					extent = max(extent, thisAtm.cartX.norm2());  // update fragment extent
					mass  += thisAtm.elt;                         //    and mass

					atoms[pos][thisAtm.atmid] = thisAtm;

					if (thisResId == resids[Noffset] ) {   // make sure no waters get mixed in
						center_res.push_back(thisAtm);
					}
				} else {
					if(radius != 0){
						thisAtm.parseLine(line);
						thisAtm.cartX = thisAtm.cartX - centerCA;  // center fragment

						if (thisAtm.cartX.norm2() <= radius2) {
							atoms_sphere.push_back(thisAtm);
							//cout << thisAtm.to_string() << endl;
						}
					} else {
						if (pos > merLength){
							break;
						}
					}
				}
			}
		}
		extent = sqrt(extent);
	}

	//
	// read an nmer contained w/i a PDB file
	void readFile(string pdbfile, char tgtChain, int tgtResid, double radius = 0.0) {
		this->pdbfile = pdbfile;
		this->tgtChain = tgtChain;
		this->tgtResid = tgtResid;

		atoms.clear(); resids.clear(); atoms_sphere.clear(); center_res.clear();
		mass = extent = 0.0;

		double radius2 = square(radius+1.0);

		atoms.resize(merLength);
		resids.resize(merLength, RESID_NONE);
		resTypes.resize(merLength);

		ifstream pdbin(pdbfile.c_str());
		if (!pdbin.is_open()) {
			cerr << "ERROR: Unable to open PDB file '" << pdbfile << "'\n";
			throw 0;
		}

		vec3<double> centerCA;
		streampos fragStart=0;
		Atom thisAtm;
		int thisResId=-1, lastResId, breakResId;
		int Noffset = (merLength-1)/2;
		string line;
		bool fragRead = false;
		char lastChain='\0', thisChain;
		int resid_counter = -10;

		//
		// PASS 1
		//     find resids of included residues
		//     find coords of center CA
		//
		while(getline(pdbin,line)) {

			if (    (line.substr(0,6) == "ATOM  " || line.substr(0,6) == "HETATM")     // atom entry
			     && (tgtChain == '0' || tgtChain == line[21])                          // correct chain
			     && (line[16]=='A' || line[16]==' ')                                   // correct conformation
			     && (line.substr(13,2) == "CA")                                        // alpha carbon
			   ) {
				thisChain = line[21];
				if (thisChain != lastChain) {
					resid_counter += 10;   // don't connect adj. chains
					lastChain = thisChain;
				}

				if ( resid_counter == tgtResid - Noffset - 1) {
					if (radius == 0.0) 	fragStart = pdbin.tellg();
				} else if (    resid_counter >= tgtResid - Noffset
				            && resid_counter <= tgtResid + Noffset ) {
					//cout << line << endl;

					fragRead = true;
					thisResId = atoi(line.substr(22,4));
					resids[resid_counter - tgtResid + Noffset] = thisResId ;
					resTypes[resid_counter - tgtResid + Noffset] = line.substr(17,3);

					if ( (resid_counter == tgtResid)  ) {
						centerCA = vec3<double>(atof(line.substr(30,8)),
												atof(line.substr(38,8)),
												atof(line.substr(46,8)));
						tgtChain = thisChain;
					}

					if ( resid_counter == tgtResid + Noffset)
						break;
				}

				++resid_counter;
			} else if (line.substr(0,6) == "TER   " && fragRead) {
				break;
			}

		}
		//cerr << endl;

		breakResId=resids[0]; for (int i=1; i<merLength; i++) breakResId=max(breakResId, resids[i]);
		//
		// PASS 2
		//     read in included residues
		//     translate coordinates so center CA is at the origin
		//
		pdbin.clear();   // clear eof()
		pdbin.seekg(fragStart, ios::beg);
		fragRead = false;

		while(getline(pdbin,line)) {
			if (     (line.substr(0,6) == "ATOM  " || line.substr(0,6) == "HETATM")
			     && (tgtChain == line[21])
			     && ( line[16]=='A' || line[16]==' ' )
				 &&   line[13] != 'H' ) {
				thisResId = atoi(line.substr(22,4));


				// 5-mer
				vector<int>::iterator p = find(resids.begin(), resids.end(), thisResId);
				//vector<int>::iterator p = resids.find( thisResId );
				if (p != resids.end() && line.substr(17,3) != "HOH") {
					//cout << line << endl;

					fragRead = true;
					int idx = distance(resids.begin(),p);

					thisAtm.parseLine(line);
					thisAtm.cartX = thisAtm.cartX - centerCA;  // center fragment

					if (radius != 0.0 && thisAtm.cartX.norm2() <= radius2) {
						atoms_sphere.push_back(thisAtm);
						//cout << thisAtm.to_string() << endl;
					}

					extent = max(extent, thisAtm.cartX.norm2());  // update fragment extent
					mass  += thisAtm.elt;                         //    and mass

					atoms[idx][thisAtm.atmid] = thisAtm;
					//cout << "atoms[" << idx << "][" << atmid << "] = " << thisAtm.cartX.to_string() << endl;

					if (thisResId == resids[Noffset] ) {   // make sure no waters get mixed in
						center_res.push_back(thisAtm);
					}

				} else {

					if (radius != 0.0) {
						thisAtm.parseLine(line);
						thisAtm.cartX = thisAtm.cartX - centerCA;  // center fragment

						if (thisAtm.cartX.norm2() <= radius2) {
							atoms_sphere.push_back(thisAtm);
							//cout << thisAtm.to_string() << endl;
						}
					} else {
						if (thisResId > breakResId) break;
					}

				}
			} else {
				if (line.substr(0,6) == "TER   " && radius == 0.0 && fragRead) {
					break;
				}
			}
		}
		//cout << "    ++ Read " << atoms_sphere.size() << " atoms within " << radius << " A\n";

		extent = sqrt(extent);
	}

	//
	// set to 1 all points within 'dist' of an atom in 'coords'
	void getMask(map_3d<double> &mask, Crystal& crystInfo, double dist, map_3d<bool> *interpMask = NULL) {
		double dist2       = dist*dist;
		double maskExtent2 = square(dist+extent);
		vec3<int> dims = mask.getDims();
		vec3<double> fX,del_fX;   // frac coords

		vector< vec3<double> > f_coords;

		// compute fractional coords of the fragment
		map< string , Atom >::iterator p;
		for (int i=0; i<merLength; i++) {
			for (p=atoms[i].begin(); p != atoms[i].end(); ++p) {
				f_coords.push_back(p->second.cartX.apply(crystInfo.c2f));
				f_coords.back().mod(1);
			}
		}
		int N = f_coords.size();

		mask.setAll(0.0);
		if (interpMask) interpMask->setAll(false);

		for (int i=0; i<dims[0]; i++) {
			fX[0] = (double)i/dims[0];
			for (int j=0; j<dims[1]; j++) {
				fX[1] = (double)j/dims[1];
				for (int k=0; k<dims[2]; k++) {
					fX[2] = (double)k/dims[2];

					fX.minmod1();

					// if we're beyond the extent of the fragment, no need to check each atom
					if (fX.apply(crystInfo.f2c).norm2() < maskExtent2) {
						if (interpMask) interpMask->set(i,j,k,true);

						// otherwise, check each atom in the fragment ...
						for (int atm=0; atm<N; atm++) {
							del_fX = fX-f_coords[atm]; del_fX.minmod1();

							if (del_fX.apply(crystInfo.f2c).norm2() < dist2) {
								mask.set(i,j,k,1.0);
								break;   // no need to look at other atoms
							}
						}
					}

				}
			}
		}
	}


	//
	void getMask(map_3d<double> &mask, Crystal& crystInfo, double dist, double radius_max) {
		double dist2       = dist*dist;
		double radius_max2 = square(radius_max + 1.0);
		vec3<int> dims = mask.getDims();
		vec3<double> fX,del_fX;   // frac coords

		vector< vec3<double> > f_coords;

		// compute fractional coords of the fragment
		map< string , Atom >::iterator p;
		for (int i=0; i<merLength; i++) {
			for (p=atoms[i].begin(); p != atoms[i].end(); ++p) {
				//if (  i==(merLength-1)/2
				//   || p->first == " C  " || p->first == " CA "
				//   || p->first == " O  " || p->first == " N  "
				//   || p->first == " CB " ) {
					f_coords.push_back(p->second.cartX.apply(crystInfo.c2f));
					f_coords.back().mod(1);
				//}
			}
		}
		int N = f_coords.size();

		mask.setAll(0.0);

		// loop thru points
		for (int i=0; i<dims[0]; i++) {
			fX[0] = (double)i/dims[0];
			for (int j=0; j<dims[1]; j++) {
				fX[1] = (double)j/dims[1];
				for (int k=0; k<dims[2]; k++) {
					fX[2] = (double)k/dims[2];

					fX.minmod1();

					// if we're beyond the extent of the fragment, no need to check each atom
					if (fX.apply(crystInfo.f2c).norm2() <= radius_max2) {

						// otherwise, check each atom in the fragment ...
						for (int atm=0; atm<N; atm++) {
							del_fX = fX-f_coords[atm]; del_fX.minmod1();

							if (del_fX.apply(crystInfo.f2c).norm2() < dist2) {
								mask.set(i,j,k,1.0);
								break;   // no need to look at other atoms
							}
						}
					}

				}
			}
		}
	}

	//
	// compute the electron density of this fragment
	//
	void getFc(map_3d<double> &rho, Crystal &crystInfo, double resolution, bool ALLATOM = false) {
		vec3<int> dims = rho.getDims();
		rho.reset();
		int nreses = atoms.size(), natoms, thisElt;

		map_3d< complex<double> > Frho(dims, complex<double>(0,0) ),   // calculated diffraction
		                          rho_cplx(dims);
		vector< map_3d<double> > f0(NUM_ELEMENTS);     // scattering factors for each element
		map_3d<double> *thisf0;
		map_3d<double> mask(dims,-1.0);                 // mask diffraction pattern by resolution
		double a1,a2,a3,a4,b1,b2,b3,b4,c;
		double S,S2, thisB;
		int H,K,L;


		// resolution mask
		double reso_cutoff = 1.0/square(resolution);
		for (int h=0; h<dims[0]; h++) {
			H = h<(dims[0]/2) ? h : h - dims[0];

			for (int k=0; k<dims[1]; k++) {
				K = k<(dims[1]/2) ? k : k - dims[1];

				for (int l=0; l<(dims[2]+1)/2; l++) {
					L = l; //l<(dims[2]/2) ? l : l - dims[2];

					S2 = crystInfo.S(H,K,L);
					if (S2 <= reso_cutoff) {
						mask.set(h,k,l, S2);
					}
				}
			}
		}


		map< string , Atom >::iterator p;
		for (int i=0; i<merLength; i++) {
			for (p=atoms[i].begin(); p != atoms[i].end(); ++p) {
				//  get fractional coords
				vec3<double> fracX = p->second.cartX.apply(crystInfo.c2f);
				fracX.mod(1);

				thisElt = p->second.elt - 1;
				if (thisElt < 0) continue;

				thisf0 = &f0[thisElt];
				thisB = p->second.B;

				// compute scattering factors (if necessary)
				if (thisf0->getDimsProd() == 0) {
					thisf0->alloc(dims);
					thisf0->setAll(0);

					// kromer-mann coeffs
					c = diffraction[thisElt][0];
					a1 = diffraction[thisElt][1];
					a2 = diffraction[thisElt][2];
					a3 = diffraction[thisElt][3];
					a4 = diffraction[thisElt][4];
					b1 = diffraction[thisElt][5];
					b2 = diffraction[thisElt][6];
					b3 = diffraction[thisElt][7];
					b4 = diffraction[thisElt][8];

					for (int h=0; h<dims[0]; h++) {
						H = h<(dims[0]/2) ? h : h - dims[0];

						for (int k=0; k<dims[1]; k++) {
							K = k<(dims[1]/2) ? k : k - dims[1];

							for (int l=0; l<(dims[2]+1)/2; l++) {
								L = l; //l<(dims[2]/2) ? l : l - dims[2];

								S2 = mask.get(h,k,l);
								if (S2 > 0.0) {
									thisf0->set(h,k,l, c +
														   a1 * exp(-b1*S2/4) +
														   a2 * exp(-b2*S2/4) +
														   a3 * exp(-b3*S2/4) +
														   a4 * exp(-b4*S2/4) );
								}
							}
						}
					}
				}


				// compute Fc (the calculated diffraction)
				double dotProd, Bscale;
				for (int h=0; h<dims[0]; h++) {
					H = h<(dims[0]/2) ? h : h - dims[0];

					for (int k=0; k<dims[1]; k++) {
						K = k<(dims[1]/2) ? k : k - dims[1];

						for (int l=0; l<(dims[2]+1)/2; l++) {
							L = l; //l<(dims[2]/2) ? l : l - dims[2];

							// inner loop
							S2 = mask.get(h,k,l);
							if (S2 > 0.0) {
								dotProd = -(fracX[0]*H + fracX[1]*K + fracX[2]*L);
								Bscale  = thisB<1e-6? 1 : exp(- thisB * S2/4);
								Bscale *= thisf0->get(h,k,l);
								Frho.inc(h,k,l,
				                         complex<double>(Bscale * cos(2*M_PI*dotProd),
				                                         Bscale * sin(2*M_PI*dotProd) ) );
							}
						}
					}
				}
			}
		}

		// reflect
		Frho.set(0,0,0, complex<double>(mass,0));
		for (int h=0; h<dims[0]; h++)
			for (int k=0; k<dims[1]; k++)
				for (int l=1; l<(dims[2]+1)/2; l++) {
					Frho.set( h==0? 0 : dims[0]-h,
							  k==0? 0 : dims[1]-k,
							  dims[2]-l,
							  conj(Frho.get(h,k,l)));
				}

		// inverse fft  (no need to scale, scaling handled in match)
		ifft( Frho, rho_cplx );
		for (int i=0; i<dims[0]*dims[1]*dims[2]; i++)
			rho.set(i, rho_cplx.get(i).real() );
	}


	//
	// compute the electron density of this fragment
	//
	void getFcSphere(map_3d<double> &rho, Crystal &crystInfo, double resolution, bool ALLATOM = false) {
		vec3<int> dims = rho.getDims();
		rho.reset();
		int nreses = atoms.size(), natoms, thisElt;

		static int firstCall = 1;

		map_3d< complex<double> > Frho(dims, complex<double>(0,0) ),   // calculated diffraction
		                          rho_cplx(dims);
		vector< map_3d<double> > f0(NUM_ELEMENTS);     // scattering factors for each element
		map_3d<double> *thisf0;
		map_3d<double> mask(dims,-1.0);                 // mask diffraction pattern by resolution
		double a1,a2,a3,a4,b1,b2,b3,b4,c;
		double S,S2, thisB;
		int H,K,L;


		// resolution mask
		double reso_cutoff = 1.0/square(resolution);
		for (int h=0; h<dims[0]; h++) {
			H = h<(dims[0]/2) ? h : h - dims[0];

			for (int k=0; k<dims[1]; k++) {
				K = k<(dims[1]/2) ? k : k - dims[1];

				for (int l=0; l<(dims[2]+1)/2; l++) {
					L = l; //l<(dims[2]/2) ? l : l - dims[2];

					S2 = crystInfo.S(H,K,L);
					if (S2 <= reso_cutoff) {
						mask.set(h,k,l, S2);
					}
				}
			}
		}


		for (int i=0; i<atoms_sphere.size(); i++) {
			vec3<double> fracX = atoms_sphere[i].cartX.apply(crystInfo.c2f);
			fracX.mod(1);
			thisElt = atoms_sphere[i].elt - 1;
			thisf0 = &f0[thisElt];
			thisB = atoms_sphere[i].B;

			// compute scattering factors (if necessary)
			if (thisf0->getDimsProd() == 0) {
				thisf0->alloc(dims);
				thisf0->setAll(0);

				// kromer-mann coeffs
				c = diffraction[thisElt][0];
				a1 = diffraction[thisElt][1];
				a2 = diffraction[thisElt][2];
				a3 = diffraction[thisElt][3];
				a4 = diffraction[thisElt][4];
				b1 = diffraction[thisElt][5];
				b2 = diffraction[thisElt][6];
				b3 = diffraction[thisElt][7];
				b4 = diffraction[thisElt][8];

				for (int h=0; h<dims[0]; h++) {
					H = h<(dims[0]/2) ? h : h - dims[0];

					for (int k=0; k<dims[1]; k++) {
						K = k<(dims[1]/2) ? k : k - dims[1];

						for (int l=0; l<(dims[2]+1)/2; l++) {
							L = l; //l<(dims[2]/2) ? l : l - dims[2];

							S2 = mask.get(h,k,l);
							if (S2 > 0.0) {
								thisf0->set(h,k,l, c +
														a1 * exp(-b1*S2/4) +
														a2 * exp(-b2*S2/4) +
														a3 * exp(-b3*S2/4) +
														a4 * exp(-b4*S2/4) );
							}
						}
					}
				}
			}


			// compute Fc (the calculated diffraction)
			double dotProd, Bscale;
			for (int h=0; h<dims[0]; h++) {
				H = h<(dims[0]/2) ? h : h - dims[0];

				for (int k=0; k<dims[1]; k++) {
					K = k<(dims[1]/2) ? k : k - dims[1];

					for (int l=0; l<(dims[2]+1)/2; l++) {
						L = l; //l<(dims[2]/2) ? l : l - dims[2];

						// inner loop
						S2 = mask.get(h,k,l);
						if (S2 > 0.0) {
							dotProd = -(fracX[0]*H + fracX[1]*K + fracX[2]*L);
							Bscale  = thisB < 1e-6? 1 : exp(- thisB * S2/4);
							Bscale *= thisf0->get(h,k,l);
							Frho.inc(h,k,l,
										complex<double>(Bscale * cos(2*M_PI*dotProd),
														Bscale * sin(2*M_PI*dotProd) ) );
						}
					}
				}
			}
		}

		// reflect
		Frho.set(0,0,0, complex<double>(mass,0));
		for (int h=0; h<dims[0]; h++)
			for (int k=0; k<dims[1]; k++)
				for (int l=1; l<(dims[2]+1)/2; l++) {
					Frho.set( h==0? 0 : dims[0]-h,
							  k==0? 0 : dims[1]-k,
							  dims[2]-l,
							  conj(Frho.get(h,k,l)));
				}

		// inverse fft  (no need to scale, scaling handled in match)
		ifft( Frho, rho_cplx );
		for (int i=0; i<dims[0]*dims[1]*dims[2]; i++)
			rho.set(i, rho_cplx.get(i).real() );
	}


	//
	// rotate an nmer
	//     leave 'atoms_sphere' unrotated
	void rotate(mat33<double> R) {
		map< string , Atom >::iterator p;
		for (int i=0; i<merLength; i++)
			for (p=atoms[i].begin(); p != atoms[i].end(); ++p)
				p->second.cartX = p->second.cartX.apply(R);

		for (int i=0; i<center_res.size(); ++i) {
			center_res[i].cartX = center_res[i].cartX.apply(R);
		}
	}

	//
	// translate an nmer
	//       leave 'atoms_sphere' untranslated
	void translate(vec3<double> T) {
		map< string , Atom >::iterator p;
		for (int i=0; i<merLength; i++)
			for (p=atoms[i].begin(); p != atoms[i].end(); ++p)
				p->second.cartX = p->second.cartX + T;
		for (int i=0; i<center_res.size(); ++i) {
			center_res[i].cartX = center_res[i].cartX + T;
		}
	}

	//
	// quick check of legality
	//    make sure each res has 1 and only 1 CA
	//    make sure CA's are 3.6 - 4.0 A apart
	bool legalStructure( string mer ) {
		double dist = 0;

		// check that 'center_res' was correctly parsed
		int ca_count = 0;
		int ctr_count = center_res.size();

		if (ctr_count < 4) return false;

		for (int i=0; i<ctr_count; ++i)
			if ( center_res[i].atmid == " CA ") ca_count++;
		if (ca_count != 1) return false;

		// check for seq-str agreement
		for (int i=0; i<merLength; i++) {
			if (resids[i] == RESID_NONE && mer[i] != '_') return false;
		}

		// check for intact backbone
		for (int i=0; i<merLength; i++) {
			if (resids[i] != RESID_NONE) {
				if ( atoms[i].count(" N  ") != 1 ) return false;
				if ( atoms[i].count(" CA ") != 1 ) return false;
				if ( atoms[i].count(" C  ") != 1 ) return false;
				if ( atoms[i].count(" O  ") != 1 ) return false;
			}
		}

		// check for illegal dists
		for (int i=0; i<merLength-1; i++) {
			if (resids[i] != RESID_NONE && resids[i+1] != RESID_NONE) {
				dist = (getCA(i) - getCA(i+1)).norm2();
				if (dist > 4.0*4.0 || dist < 3.6*3.6) return false;
			}
		}
		return true;
	}

	// debugging
	string to_string() {
		ostringstream out;
		int atmCounter = 0;
		for (int i=0; i<merLength; i++) {
			out << setiosflags(ios_base::right|ios::fixed) << setprecision(3);

			for (  map< string , Atom >::iterator pX = atoms[i].begin();
			       pX != atoms[i].end();
			       ++pX) {
				out << "ATOM  " << setw(5) << atmCounter++ << " " << pX->first << " "
					<< resTypes[i] << " " << (char)('A') << setw(4) << i+1 << "    ";
				out << setw(8) << pX->second.cartX[0]
					<< setw(8) << pX->second.cartX[1]
					<< setw(8) << pX->second.cartX[2];
				out << "  1.00 0.000" << endl;
			}
		}
		return out.str();
	}

	// get the middle residue
	void getCenterRes(Residue &thisRes) {
		thisRes.atoms = center_res;
	}

	string getCenterRestype() {
		return resTypes[(merLength-1)/2];
	}
	string getMer(){
		string mer = "";
		for(int i = 0;i < 5;i++){
			mer += oneLetterCode(resTypes[i]);
		}
		return mer;
	}
	vec3<double> getCA(int i) {
		if (atoms[i].find(" CA ") != atoms[i].end())
			return atoms[i][" CA "].cartX;
		return vec3<double>(0,0,0);
	}
	vec3<double> getAtom(int i, string id) {
		if (atoms[i].find(id) != atoms[i].end())
			return atoms[i][id].cartX;
		//else
		//	cerr << "WARNING: atom " << id << " not found\n";

		return vec3<double>(0,0,0);
	}

	inline int nres() { return merLength; };

	void averageBfactors( vector<Nmer> &data, vector<int> exs) {
		map< string , Atom >::iterator pX,pY;

		for (int i=0; i<merLength; i++) {
			// for each atom ...
			for (pX = atoms[i].begin(); pX != atoms[i].end(); ++pX) {
				int nBs = 0;
				double sumB = 0.0;
				for (int j=0; j<exs.size(); j++) {
					Nmer &Y = data[ exs[j] ];
					pY = Y.atoms[i].find(pX->first);
					if (pY != Y.atoms[i].end()) {
						sumB += pY->second.B;
						nBs++;
					}
				}
				pX->second.B = sumB / nBs;
			}
		}
	}

	void clearBfactors() {
		map< string , Atom >::iterator pX;
		for (int i=0; i<merLength; i++) {
			for (pX = atoms[i].begin(); pX != atoms[i].end(); ++pX) {
				pX->second.B = 0;
			}
		}
	}

	double getExtent() {
		return extent;
	}

	//
	string pdbfile;
	char tgtChain;
	int tgtResid;

private:
	vector< map<string,Atom> > atoms;
	vector< int > resids;
	vector< string > resTypes;

	vector< Atom > atoms_sphere;  // all atoms w/i k Ang. of center
	vector< Atom > center_res;

	double extent;  // the furthest an atom lies from the origin
	double mass;    // atomic mass of the fragment

	int merLength;

	static const int RESID_NONE; // impossible for a residue to have this ID
};

const int Nmer::RESID_NONE = -999999; // impossible for a residue to have this ID


///
/// align ()
///     align fragment with vector of CAs
///
void align(vector< vec3<double> > CAs, Nmer &X, mat33<double> &rotate, double &rmsd) {
	if (X.nres() != CAs.size()) {
		cerr << "ERROR: align(X,CAtrace) X and CAtrace have an unequal number of residues ("
		     << X.nres() << " vs " << CAs.size() << ")\n";
		return;
	}

	// allocate space for call to align
	double (*xcoords)[3] = new double[5][3];
	double (*ycoords)[3] = new double[5][3];
	double R[3][3];

	int atmidx = 0;
	for (int i=0; i<5; i++) {
		if (X.resids[i] == Nmer::RESID_NONE || CAs[i].norm2() < 1e-6) continue;

		vec3<double> CAx = X.getCA(i);
		xcoords[atmidx][0] = CAx[0];
		xcoords[atmidx][1] = CAx[1];
		xcoords[atmidx][2] = CAx[2];

		vec3<double> CAy = CAs[i];
		ycoords[atmidx][0] = CAy[0];
		ycoords[atmidx][1] = CAy[1];
		ycoords[atmidx][2] = CAy[2];

		atmidx++;
	}

	calculate_rotation_rmsd_notranslate(ycoords, xcoords, atmidx, R, &rmsd);

	rotate.set(R);

	delete[] xcoords;
	delete[] ycoords;
}

///
///  align()
///      align two fragments  ; only consider common atoms
///
void align(Nmer &X, Nmer &Y, mat33<double> &rotate, double &rmsd, bool translate=true, bool allatoms=false) {
	if (X.nres() != Y.nres()) {
		cerr << "ERROR: align(X,Y) X and Y have an unequal number of residues ("
		     << X.nres() << " vs " << Y.nres() << ")\n";
		rmsd = -1;
		return;
	}

	// count upper bound on #atoms
	int natoms = 0;
	for (int i=0; i<X.nres(); i++) natoms += min(X.atoms[i].size(), Y.atoms[i].size());
	//for (int i = (X.nres()-3)/2 ; i <= (X.nres()+1)/2; ++i) natoms += min(X.atoms[i].size(), Y.atoms[i].size());

	// allocate space for call to align
	double (*xcoords)[3] = new double[natoms][3];
	double (*ycoords)[3] = new double[natoms][3];
	double R[3][3];
	map< string , Atom >::iterator pX,pY;
	int atomidx = 0;

	if (allatoms) {
		for (int i=0; i<X.nres(); i++) {
			for (pX = X.atoms[i].begin(); pX != X.atoms[i].end(); ++pX) {
				string atmid = pX->first;
				pY = Y.atoms[i].find(atmid);
				if (pY != Y.atoms[i].end()) {
					xcoords[atomidx][0] = pX->second.cartX[0];
					xcoords[atomidx][1] = pX->second.cartX[1];
					xcoords[atomidx][2] = pX->second.cartX[2];

					ycoords[atomidx][0] = pY->second.cartX[0];
					ycoords[atomidx][1] = pY->second.cartX[1];
					ycoords[atomidx][2] = pY->second.cartX[2];

					atomidx++;
				}
			}
		}
	} else {
		for (int i = (X.nres()-3)/2 ; i <= (X.nres()+1)/2; ++i) {
			for (pX = X.atoms[i].begin(); pX != X.atoms[i].end(); ++pX) {
				string atmid = pX->first;

				if (i == (X.nres()-3)/2 && atmid != " C  " && atmid != " CA " && atmid != " O  "  && atmid != " N  " ) continue;
				if (i == (X.nres()+1)/2 && atmid != " C  " && atmid != " CA " && atmid != " O  "  && atmid != " N  " ) continue;

				pY = Y.atoms[i].find(atmid);
				if (pY != Y.atoms[i].end()) {
					xcoords[atomidx][0] = pX->second.cartX[0];
					xcoords[atomidx][1] = pX->second.cartX[1];
					xcoords[atomidx][2] = pX->second.cartX[2];

					ycoords[atomidx][0] = pY->second.cartX[0];
					ycoords[atomidx][1] = pY->second.cartX[1];
					ycoords[atomidx][2] = pY->second.cartX[2];

					atomidx++;
				}
			}
		}
	}


	if (translate) {
		double mov_com[3], mov_to_ref[3];
		calculate_rotation_rmsd(xcoords, ycoords, atomidx, mov_com, mov_to_ref, R, &rmsd);
	} else {
		calculate_rotation_rmsd_notranslate(xcoords, ycoords, atomidx, R, &rmsd);
	}
	rotate.set(R);  // don't use translation

	delete[] xcoords;
	delete[] ycoords;
}



///
///  calcRMSd()
///      the rmsd of two fragments ; only consider common atoms
///
void calcRMSd(Nmer &X, Nmer &Y, double &rmsd) {
	if (X.nres() != Y.nres()) {
		cerr << "ERROR: align(X,Y) X and Y have an unequal number of residues ("
		     << X.nres() << " vs " << Y.nres() << ")\n";
		rmsd = -1;
		return;
	}

	map< string , Atom >::iterator pX,pY;
	int natoms = 0;
	rmsd = 0.0;
	for (int i=0; i< X.nres(); i++) {
		for (pX = X.atoms[i].begin(); pX != X.atoms[i].end(); ++pX) {
			string atmid = pX->first;
			pY = Y.atoms[i].find(atmid);
			if (pY != Y.atoms[i].end()) {
				rmsd += (pX->second.cartX - pY->second.cartX).norm2();
				natoms++;
			}
		}
	}
	rmsd = sqrt(rmsd/natoms);
}



/////////////////////////
/////////////////////////
///
///   SELENIUM MAP
///
class PriorAtomList {
public:
	void load(string pdbfile) {
		nSEs = 0;
		ifstream pdbin(pdbfile.c_str());
		if (!pdbin.is_open()) {
			cerr << "ERROR: Unable to open Prior Atom PDB file '" << pdbfile << "'\n";
			throw 0;
		}

		string line;
		Atom thisAtom;

		while(getline(pdbin,line)) {
			if (    (line.substr(0,6) == "ATOM  " || line.substr(0,6) == "HETATM")
				 && (line.substr(12,2) == "SE" || line.substr(13,2) == "SE" || line.substr(12,2) == "Se" || line.substr(13,2) == "Se" ||
				     line.substr(13,2) == "CA" || line.substr(13,2) == "CA")){
				thisAtom.parseLine(line);
				PriorAtoms.push_back(thisAtom);
				if(thisAtom.elt == 34){
					nSEs++;
				}
			}
		}
	}

	inline int nAtoms() { return PriorAtoms.size(); }
	inline Atom& operator[](int i) { return PriorAtoms[i]; }
	inline const Atom& operator[](int i) const { return PriorAtoms[i]; }
	inline int nSe() { return nSEs; }
	inline int nCA() { return nAtoms() - nSEs; }

private:
	vector< Atom > PriorAtoms;
	int nSEs;
};

#endif

